@@ -1,6 +1,5 @@
 {-# LANGUAGE ExistentialQuantification #-}
 {-# LANGUAGE LambdaCase #-}
-{-# LANGUAGE OverloadedStrings #-}
 {-# LANGUAGE RecordWildCards #-}
 
 module Network.TLS.Context.Internal (
@@ -58,8 +57,6 @@ module Network.TLS.Context.Internal (
     restoreHState,
     getStateRNG,
     tls13orLater,
-    addCertRequest13,
-    getCertRequest13,
     decideRecordVersion,
 
     -- * Misc
@@ -81,6 +78,11 @@ module Network.TLS.Context.Internal (
     getPeerRecordLimit,
     checkPeerRecordLimit,
     newRecordLimitRef,
+
+    -- * ECH
+    HPKEF,
+    getTLS13HPKE,
+    setTLS13HPKE,
 ) where
 
 import Control.Concurrent.MVar
@@ -117,7 +119,7 @@ data Context
     -- ^ return the backend object associated with this context
     , ctxSupported :: Supported
     , ctxShared :: Shared
-    , ctxLimit :: Limit
+    , ctxDebug :: DebugParams
     , ctxTLSState :: MVar TLSState
     , ctxMeasurement :: IORef Measurement
     , ctxEOF_ :: IORef Bool
@@ -133,13 +135,11 @@ data Context
     , ctxRoleParams :: RoleParams
     -- ^ hooks for this context
     , ctxLocks :: Locks
-    , ctxKeyLogger :: String -> IO ()
     , ctxHooks :: IORef Hooks
     , -- TLS 1.3
       ctxTLS13State :: IORef TLS13State
     , ctxPendingRecvActions :: IORef [PendingRecvAction]
     , ctxPendingSendAction :: IORef (Maybe (Context -> IO ()))
-    , ctxCertRequests :: IORef [Handshake13]
     -- ^ pending post handshake authentication requests
     , -- QUIC
       ctxRecordLayer :: RecordLayer a
@@ -152,8 +152,11 @@ data Context
     -- ^ maximum size of plaintext fragments, val + 1 is used for TLS 1.3
     , ctxPeerRecordLimit :: IORef RecordLimit
     -- ^ maximum size of plaintext fragments, val + 1 is used for TLS 1.3
+    , ctxHPKE :: IORef (Maybe (HPKEF, Int))
     }
 
+type HPKEF = ByteString -> ByteString -> IO ByteString
+
 data RecordLimit
     = NoRecordLimit -- for QUIC
     | RecordLimit
@@ -268,10 +271,11 @@ data Established
 
 data PendingRecvAction
     = -- | simple pending action. The first 'Bool' is necessity of alignment.
-      PendingRecvAction Bool (Handshake13 -> IO ())
+      -- The second bool is update transcript hash.
+      PendingRecvAction Bool Bool (Handshake13 -> IO ())
     | -- | pending action taking transcript hash up to preceding message
       --   The first 'Bool' is necessity of alignment.
-      PendingRecvActionHash Bool (ByteString -> Handshake13 -> IO ())
+      PendingRecvActionHash Bool (TranscriptHash -> Handshake13 -> IO ())
 
 updateMeasure :: Context -> (Measurement -> Measurement) -> IO ()
 updateMeasure ctx = modifyIORef' (ctxMeasurement ctx)
@@ -292,7 +296,7 @@ contextGetInformation :: Context -> IO (Maybe Information)
 contextGetInformation ctx = do
     ver <- usingState_ ctx $ gets stVersion
     hstate <- getHState ctx
-    let (ms, ems, cr, sr, hm13, grp) =
+    let (ms, ems, cr, sr, hm13, grp, ech) =
             case hstate of
                 Just st ->
                     ( hstMainSecret st
@@ -301,8 +305,9 @@ contextGetInformation ctx = do
                     , hstServerRandom st
                     , if ver == Just TLS13 then Just (hstTLS13HandshakeMode st) else Nothing
                     , hstSupportedGroup st
+                    , hstTLS13ECHAccepted st
                     )
-                Nothing -> (Nothing, False, Nothing, Nothing, Nothing, Nothing)
+                Nothing -> (Nothing, False, Nothing, Nothing, Nothing, Nothing, False)
     (cipher, comp) <-
         readMVar (ctxRxRecordState ctx) <&> \st -> (stCipher st, stCompression st)
     let accepted = case hstate of
@@ -325,6 +330,7 @@ contextGetInformation ctx = do
                         , infoTLS12Resumption = tls12resumption
                         , infoTLS13HandshakeMode = hm13
                         , infoIsEarlyDataAccepted = accepted
+                        , infoIsECHAccepted = ech
                         }
         _ -> return Nothing
 
@@ -348,7 +354,7 @@ ctxWithHooks :: Context -> (Hooks -> IO a) -> IO a
 ctxWithHooks ctx f = readIORef (ctxHooks ctx) >>= f
 
 contextModifyHooks :: Context -> (Hooks -> Hooks) -> IO ()
-contextModifyHooks ctx = modifyIORef (ctxHooks ctx)
+contextModifyHooks ctx = modifyIORef' (ctxHooks ctx)
 
 setEstablished :: Context -> Established -> IO ()
 setEstablished ctx = writeIORef (ctxEstablished_ ctx)
@@ -456,31 +462,16 @@ tls13orLater ctx = do
         Left _ -> False
         Right v -> v >= TLS13
 
-addCertRequest13 :: Context -> Handshake13 -> IO ()
-addCertRequest13 ctx certReq = modifyIORef (ctxCertRequests ctx) (certReq :)
-
-getCertRequest13 :: Context -> CertReqContext -> IO (Maybe Handshake13)
-getCertRequest13 ctx context = do
-    let ref = ctxCertRequests ctx
-    l <- readIORef ref
-    let (matched, others) = partition (\cr -> context == fromCertRequest13 cr) l
-    case matched of
-        [] -> return Nothing
-        (certReq : _) -> writeIORef ref others >> return (Just certReq)
-  where
-    fromCertRequest13 (CertRequest13 c _) = c
-    fromCertRequest13 _ = error "fromCertRequest13"
-
 --------------------------------
 
 setMyRecordLimit :: Context -> Maybe Int -> IO ()
-setMyRecordLimit ctx msiz = modifyIORef (ctxMyRecordLimit ctx) change
+setMyRecordLimit ctx msiz = modifyIORef' (ctxMyRecordLimit ctx) change
   where
     change (RecordLimit n _) = RecordLimit n msiz
     change x = x
 
 enableMyRecordLimit :: Context -> IO ()
-enableMyRecordLimit ctx = modifyIORef (ctxMyRecordLimit ctx) change
+enableMyRecordLimit ctx = modifyIORef' (ctxMyRecordLimit ctx) change
   where
     change (RecordLimit _ (Just n)) = RecordLimit n Nothing
     change x = x
@@ -500,13 +491,13 @@ checkMyRecordLimit ctx = chk <$> readIORef (ctxMyRecordLimit ctx)
 --------------------------------
 
 setPeerRecordLimit :: Context -> Maybe Int -> IO ()
-setPeerRecordLimit ctx msiz = modifyIORef (ctxPeerRecordLimit ctx) change
+setPeerRecordLimit ctx msiz = modifyIORef' (ctxPeerRecordLimit ctx) change
   where
     change (RecordLimit n _) = RecordLimit n msiz
     change x = x
 
 enablePeerRecordLimit :: Context -> IO ()
-enablePeerRecordLimit ctx = modifyIORef (ctxPeerRecordLimit ctx) change
+enablePeerRecordLimit ctx = modifyIORef' (ctxPeerRecordLimit ctx) change
   where
     change (RecordLimit _ (Just n)) = RecordLimit n Nothing
     change x = x
@@ -526,3 +517,11 @@ checkPeerRecordLimit ctx = chk <$> readIORef (ctxPeerRecordLimit ctx)
 newRecordLimitRef :: Maybe Int -> IO (IORef RecordLimit)
 newRecordLimitRef Nothing = newIORef NoRecordLimit
 newRecordLimitRef (Just n) = newIORef $ RecordLimit n Nothing
+
+--------------------------------
+
+setTLS13HPKE :: Context -> HPKEF -> Int -> IO ()
+setTLS13HPKE ctx func nenc = writeIORef (ctxHPKE ctx) $ Just (func, nenc)
+
+getTLS13HPKE :: Context -> IO (Maybe (HPKEF, Int))
+getTLS13HPKE ctx = readIORef $ ctxHPKE ctx
