@@ -1,4 +1,5 @@
 {-# LANGUAGE OverloadedStrings #-}
+{-# LANGUAGE RecordWildCards #-}
 {-# LANGUAGE ScopedTypeVariables #-}
 {-# OPTIONS_HADDOCK hide #-}
 
@@ -34,19 +35,17 @@ import qualified Data.ByteString.Lazy as L
 import Data.IORef
 import System.Timeout
 
-import Network.TLS.Cipher
 import Network.TLS.Context
-import Network.TLS.Crypto
 import Network.TLS.Extension
 import Network.TLS.Handshake
 import Network.TLS.Handshake.Common
 import Network.TLS.Handshake.Common13
-import Network.TLS.Handshake.Process
+import Network.TLS.Handshake.Server
 import Network.TLS.Handshake.State
 import Network.TLS.Handshake.State13
+import Network.TLS.Handshake.TranscriptHash
 import Network.TLS.IO
 import Network.TLS.Imports
-import Network.TLS.KeySchedule
 import Network.TLS.Parameters
 import Network.TLS.PostHandshake
 import Network.TLS.Session
@@ -55,8 +54,6 @@ import qualified Network.TLS.State as S
 import Network.TLS.Struct
 import Network.TLS.Struct13
 import Network.TLS.Types (
-    AnyTrafficSecret (..),
-    ApplicationSecret,
     HostName,
     Role (..),
  )
@@ -148,7 +145,7 @@ bye_ ctx = liftIO $ do
 
 -- | If the ALPN extensions have been used, this will
 -- return get the protocol agreed upon.
-getNegotiatedProtocol :: MonadIO m => Context -> m (Maybe B.ByteString)
+getNegotiatedProtocol :: MonadIO m => Context -> m (Maybe ByteString)
 getNegotiatedProtocol ctx = liftIO $ usingState_ ctx S.getNegotiatedProtocol
 
 -- | If the Server Name Indication extension has been used, return the
@@ -196,7 +193,7 @@ sendData ctx dataToSend = liftIO $ do
 
 -- | Get data out of Data packet, and automatically renegotiate if a Handshake
 -- ClientHello is received.  An empty result means EOF.
-recvData :: MonadIO m => Context -> m B.ByteString
+recvData :: MonadIO m => Context -> m ByteString
 recvData ctx = liftIO $ do
     tls13 <- tls13orLater ctx
     withReadLock ctx $ do
@@ -210,7 +207,7 @@ recvData ctx = liftIO $ do
         -- will impact the validity of the context.
         if tls13 then recvData13 ctx else recvData12 ctx
 
-recvData12 :: Context -> IO B.ByteString
+recvData12 :: Context -> IO ByteString
 recvData12 ctx = do
     pkt <- recvPacket12 ctx
     either (onError terminate12) process pkt
@@ -235,13 +232,13 @@ recvData12 ctx = do
     -- when receiving empty appdata, we just retry to get some data.
     process (AppData "") = recvData12 ctx
     process (AppData x) = return x
-    process p =
+    process p = do
         let reason = "unexpected message " ++ show p
-         in terminate12 (Error_Misc reason) AlertLevel_Fatal UnexpectedMessage reason
+        terminate12 (Error_Misc reason) AlertLevel_Fatal UnexpectedMessage reason
 
     terminate12 = terminateWithWriteLock ctx (sendPacket12 ctx . Alert)
 
-recvData13 :: Context -> IO B.ByteString
+recvData13 :: Context -> IO ByteString
 recvData13 ctx = do
     mdat <- tls13stPendingRecvData <$> getTLS13State ctx
     case mdat of
@@ -283,9 +280,9 @@ recvData13 ctx = do
                 | n > 0 -> do
                     setEstablished ctx $ EarlyDataNotAllowed (n - 1)
                     recvData13 ctx -- ignore "x"
-                | otherwise ->
+                | otherwise -> do
                     let reason = "early data deprotect overflow"
-                     in terminate13 ctx (Error_Misc reason) AlertLevel_Fatal UnexpectedMessage reason
+                    terminate13 ctx (Error_Misc reason) AlertLevel_Fatal UnexpectedMessage reason
             Established -> return x
             _ -> throwCore $ Error_Protocol "data at not-established" UnexpectedMessage
     process ChangeCipherSpec13 = do
@@ -295,18 +292,18 @@ recvData13 ctx = do
             else do
                 let reason = "CSS after Finished"
                 terminate13 ctx (Error_Misc reason) AlertLevel_Fatal UnexpectedMessage reason
-    process p =
+    process p = do
         let reason = "unexpected message " ++ show p
-         in terminate13 ctx (Error_Misc reason) AlertLevel_Fatal UnexpectedMessage reason
+        terminate13 ctx (Error_Misc reason) AlertLevel_Fatal UnexpectedMessage reason
 
     loopHandshake13 [] = return ()
     -- fixme: some implementations send multiple NST at the same time.
     -- Only the first one is used at this moment.
-    loopHandshake13 (NewSessionTicket13 life add nonce ticket exts : hs) = do
+    loopHandshake13 (NewSessionTicket13 life add nonce (SessionIDorTicket_ ticket) exts : hs) = do
         role <- usingState_ ctx S.getRole
-        unless (role == ClientRole) $
+        unless (role == ClientRole) $ do
             let reason = "Session ticket is allowed for client only"
-             in terminate13 ctx (Error_Misc reason) AlertLevel_Fatal UnexpectedMessage reason
+            terminate13 ctx (Error_Misc reason) AlertLevel_Fatal UnexpectedMessage reason
         -- This part is similar to handshake code, so protected with
         -- read+write locks (which is also what we use for all calls to the
         -- session manager).
@@ -359,18 +356,17 @@ recvData13 ctx = do
             else do
                 let reason = "received key update before established"
                 terminate13 ctx (Error_Misc reason) AlertLevel_Fatal UnexpectedMessage reason
+    -- Client only
     loopHandshake13 (h@CertRequest13{} : hs) =
         postHandshakeAuthWith ctx h >> loopHandshake13 hs
-    loopHandshake13 (h@Certificate13{} : hs) =
-        postHandshakeAuthWith ctx h >> loopHandshake13 hs
     loopHandshake13 (h : hs) = do
         rtt0 <- tls13st0RTT <$> getTLS13State ctx
         when rtt0 $ case h of
-            ServerHello13 srand _ _ _ ->
-                when (isHelloRetryRequest srand) $ do
+            ServerHello13 SH{..} ->
+                when (isHelloRetryRequest shRandom) $ do
                     clearTxRecordState ctx
                     let reason = "HRR is not allowed for 0-RTT"
-                     in terminate13 ctx (Error_Misc reason) AlertLevel_Fatal UnexpectedMessage reason
+                    terminate13 ctx (Error_Misc reason) AlertLevel_Fatal UnexpectedMessage reason
             _ -> return ()
         cont <- popAction ctx h hs
         when cont $ loopHandshake13 hs
@@ -393,11 +389,11 @@ recvHS13 ctx breakLoop = do
     loopHandshake13 [] = return ()
     -- fixme: some implementations send multiple NST at the same time.
     -- Only the first one is used at this moment.
-    loopHandshake13 (NewSessionTicket13 life add nonce ticket exts : hs) = do
+    loopHandshake13 (NewSessionTicket13 life add nonce (SessionIDorTicket_ ticket) exts : hs) = do
         role <- usingState_ ctx S.getRole
-        unless (role == ClientRole) $
+        unless (role == ClientRole) $ do
             let reason = "Session ticket is allowed for client only"
-             in terminate13 ctx (Error_Misc reason) AlertLevel_Fatal UnexpectedMessage reason
+            terminate13 ctx (Error_Misc reason) AlertLevel_Fatal UnexpectedMessage reason
         -- This part is similar to handshake code, so protected with
         -- read+write locks (which is also what we use for all calls to the
         -- session manager).
@@ -439,14 +435,14 @@ popAction ctx h hs = do
             withWriteLock ctx $
                 handleException ctx $ do
                     case action of
-                        PendingRecvAction needAligned pa -> do
+                        PendingRecvAction needAligned update pa -> do
                             when needAligned $ checkAlignment ctx hs
-                            processHandshake13 ctx h
+                            when update $ void $ updateTranscriptHash13 ctx h
                             pa h
                         PendingRecvActionHash needAligned pa -> do
                             when needAligned $ checkAlignment ctx hs
-                            d <- transcriptHash ctx
-                            processHandshake13 ctx h
+                            d <- transcriptHash ctx "Pending action"
+                            void $ updateTranscriptHash13 ctx h
                             pa d h
                     -- Client: after receiving SH, app data is coming.
                     -- this loop tries to receive it.
@@ -469,15 +465,16 @@ tryBye ctx = catchException (bye_ ctx) (\_ -> return ())
 
 onError
     :: Monad m
-    => (TLSError -> AlertLevel -> AlertDescription -> String -> m B.ByteString)
+    => (TLSError -> AlertLevel -> AlertDescription -> String -> m ByteString)
     -> TLSError
-    -> m B.ByteString
+    -> m ByteString
 onError _ Error_EOF =
     -- Not really an error.
     return B.empty
-onError terminate err =
-    let (lvl, ad) = errorToAlert err
-     in terminate err lvl ad (errorToAlertMessage err)
+onError terminate err = terminate err lvl ad reason
+  where
+    (lvl, ad) = errorToAlert err
+    reason = errorToAlertMessage err
 
 terminateWithWriteLock
     :: Context
@@ -502,6 +499,7 @@ terminateWithWriteLock ctx send err level desc reason = withWriteLock ctx $ do
                 sessionInvalidate (sharedSessionManager $ ctxShared ctx) sid
     catchException (send [(level, desc)]) (\_ -> return ())
     setEOF ctx
+    debugError (ctxDebug ctx) reason
     E.throwIO (Terminated False reason err)
 
 {-# DEPRECATED recvData' "use recvData that returns strict bytestring" #-}
@@ -509,45 +507,3 @@ terminateWithWriteLock ctx send err level desc reason = withWriteLock ctx $ do
 -- | same as recvData but returns a lazy bytestring.
 recvData' :: MonadIO m => Context -> m L.ByteString
 recvData' ctx = L.fromChunks . (: []) <$> recvData ctx
-
-keyUpdate
-    :: Context
-    -> (Context -> IO (Hash, Cipher, CryptLevel, C8.ByteString))
-    -> (Context -> Hash -> Cipher -> AnyTrafficSecret ApplicationSecret -> IO ())
-    -> IO ()
-keyUpdate ctx getState setState = do
-    (usedHash, usedCipher, level, applicationSecretN) <- getState ctx
-    unless (level == CryptApplicationSecret) $
-        throwCore $
-            Error_Protocol
-                "tried key update without application traffic secret"
-                InternalError
-    let applicationSecretN1 =
-            hkdfExpandLabel usedHash applicationSecretN "traffic upd" "" $
-                hashDigestSize usedHash
-    setState ctx usedHash usedCipher (AnyTrafficSecret applicationSecretN1)
-
--- | How to update keys in TLS 1.3
-data KeyUpdateRequest
-    = -- | Unidirectional key update
-      OneWay
-    | -- | Bidirectional key update (normal case)
-      TwoWay
-    deriving (Eq, Show)
-
--- | Updating appication traffic secrets for TLS 1.3.
---   If this API is called for TLS 1.3, 'True' is returned.
---   Otherwise, 'False' is returned.
-updateKey :: MonadIO m => Context -> KeyUpdateRequest -> m Bool
-updateKey ctx way = liftIO $ do
-    tls13 <- tls13orLater ctx
-    when tls13 $ do
-        let req = case way of
-                OneWay -> UpdateNotRequested
-                TwoWay -> UpdateRequested
-        -- Write lock wraps both actions because we don't want another packet to
-        -- be sent by another thread before the Tx state is updated.
-        withWriteLock ctx $ do
-            sendPacket13 ctx $ Handshake13 [KeyUpdate13 req]
-            keyUpdate ctx getTxRecordState setTxRecordState
-    return tls13
